package info.textgrid.middleware;

import java.io.IOException;
import java.text.ParseException;
import java.util.ArrayList;
import java.util.List;
import javax.xml.bind.JAXBElement;
import javax.xml.datatype.DatatypeConfigurationException;
import javax.xml.datatype.XMLGregorianCalendar;
import org.apache.commons.logging.LogFactory;
import info.textgrid.middleware.oaipmh.GetRecordType;
import info.textgrid.middleware.oaipmh.IdentifyType;
import info.textgrid.middleware.oaipmh.ListIdentifiersType;
import info.textgrid.middleware.oaipmh.ListMetadataFormatsType;
import info.textgrid.middleware.oaipmh.ListRecordsType;
import info.textgrid.middleware.oaipmh.ListSetsType;
import info.textgrid.middleware.oaipmh.OAIPMHType;
import info.textgrid.middleware.oaipmh.ObjectFactory;
import info.textgrid.middleware.oaipmh.RequestType;
import info.textgrid.middleware.oaipmh.VerbType;

/**
 * <p>
 * This class takes all possible arguments of an OAI-PMH request and proceeds depended on the given
 * arguments. It also deals as response handler.
 * </p>
 * 
 * @author Maximilian Brodhun, SUB Göttingen
 * @version 2019-05-28
 * @since 2014-01-29
 */
public class OAIPMHImpl implements OAIPMHProducer {

  private ErrorHandler error = new ErrorHandler();
  private org.apache.commons.logging.Log log = LogFactory.getLog(OAIPMHImpl.class);
  private RepIdentification rep;

  private RecordListDelivererDC recordListDC;
  private RecordListDelivererIDIOM recordListIDIOM;

  private RecordDelivererDC recDelivDC;
  private RecordDelivererIDIOM recDelivIDIOM;

  private IdentifierListDelivererDC identifierListDC;
  private IdentifierListDelivererIDIOM identifierListIDIOM;

  private MetadataFormatListDeliverer metadataFormatList;
  private SetDeliverer setList;

  public boolean textgrid;
  public boolean dariah;

  ListIdentifiersType lit = new ListIdentifiersType();

  /**
   * @param rep
   * @param recDelivDC
   * @param recDelivIDIOM
   * @param recordListDC
   * @param recordListIDIOM
   * @param metadataFormatList
   * @param setList
   * @param identifierList
   * @param identifierListIDIOM
   */
  public OAIPMHImpl(RepIdentification rep, RecordDelivererDC recDelivDC,
      RecordDelivererIDIOM recDelivIDIOM, RecordListDelivererDC recordListDC,
      RecordListDelivererIDIOM recordListIDIOM, MetadataFormatListDeliverer metadataFormatList,
      SetDeliverer setList, IdentifierListDelivererDC identifierList,
      IdentifierListDelivererIDIOM identifierListIDIOM) {
    this.rep = rep;
    this.recDelivDC = recDelivDC;
    this.recDelivIDIOM = recDelivIDIOM;
    this.recordListDC = recordListDC;
    this.recordListIDIOM = recordListIDIOM;
    this.metadataFormatList = metadataFormatList;
    this.setList = setList;
    this.identifierListDC = identifierList;
    this.identifierListIDIOM = identifierListIDIOM;
  }

  /*
   * (non-Javadoc)
   *
   * @see info.textgrid.middleware.OAIPMHProducer#getRequest(java.lang.String, java.lang.String,
   * java.lang.String, java.lang.String, java.lang.String, java.lang.String, java.lang.String)
   */
  public JAXBElement<OAIPMHType> postRequest(String verb, String identifier, String metadataPrefix,
      String set, String from, String until, String resumptionToken) {
    return handleRequest(verb, identifier, metadataPrefix, set, from, until, resumptionToken);
  }

  /*
   * (non-Javadoc)
   * 
   * @see info.textgrid.middleware.OAIPMHProducer#getRequest(java.lang.String, java.lang.String,
   * java.lang.String, java.lang.String, java.lang.String, java.lang.String, java.lang.String)
   */
  public JAXBElement<OAIPMHType> getRequest(String verb, String identifier, String metadataPrefix,
      String set, String from, String until, String resumptionToken) {
    return handleRequest(verb, identifier, metadataPrefix, set, from, until, resumptionToken);
  }

  /**
   * @return
   */
  public String getVersion() {
    return OaipmhServiceVersion.BUILDNAME + "-" + OaipmhServiceVersion.VERSION + "."
        + OaipmhServiceVersion.BUILDDATE;
  }

  /**
   * checks the Identify request of correctness and response including errors in case of an
   * incorrect request
   * 
   * @param obf
   * @param oai
   * @param request
   * @return
   */
  public JAXBElement<OAIPMHType> identifyRequest(ObjectFactory obf, OAIPMHType oai,
      RequestType request) {

    List<String> errorValues = new ArrayList<String>();
    JAXBElement<OAIPMHType> oaipmhRoot = obf.createOAIPMH(oai);
    IdentifyType tgRepIdentificationRequest = new IdentifyType();

    if (this.rep.requestChecker(request)) {

      tgRepIdentificationRequest.setBaseURL(this.rep.getBaseUrl());
      tgRepIdentificationRequest.setDeletedRecord(this.rep.getDeletedRecordStatus());
      tgRepIdentificationRequest.setEarliestDatestamp(this.rep.getEarliestDatestamp());
      tgRepIdentificationRequest.setGranularity(this.rep.getGranularity());
      tgRepIdentificationRequest.setRepositoryName(this.rep.getRepositoryName());
      tgRepIdentificationRequest.setProtocolVersion(this.rep.getProtocolVersion());
      tgRepIdentificationRequest.getAdminEmail().add(this.rep.getAdminMaiAddresss());
      oai.setIdentify(tgRepIdentificationRequest);

    } else {
      if (request.getIdentifier() != null) {
        errorValues.add("identifier");
      }
      if (request.getFrom() != null) {
        errorValues.add("from");
      }
      if (request.getMetadataPrefix() != null) {
        errorValues.add("metadataPrefix");
      }
      if (request.getResumptionToken() != null) {
        errorValues.add("ResumptionToken");
      }
      if (request.getUntil() != null) {
        errorValues.add("Until");
      }
      if (request.getSet() != null) {
        errorValues.add("Set");
      }

      this.error.setError(TGConstants.OAI_BAD_ARGUMENT,
          "The request includes illegal arguments: " + errorValues);
      oai.getError().add(this.error.getError());
    }

    return oaipmhRoot;
  }

  /**
   * checks the GetRecord request of correctness and response including errors in case of an
   * incorrect request
   * 
   * @param obf
   * @param oai
   * @param request
   * @return
   */
  public JAXBElement<OAIPMHType> getRecordRequest(ObjectFactory obf, OAIPMHType oai,
      RequestType request) {

    JAXBElement<OAIPMHType> oaipmhRoot = obf.createOAIPMH(oai);

    /**
     * Check request. In GetRecord: allowed parameters: identifier, metadataPrefix, verb forbidden:
     * resumptionToken, from, till, set optional: no one
     */

    ErrorHandler requestErrors = RecordDelivererAbstract.requestChecker(request);

    if (requestErrors.getError().getCode() != null) {
      oai.getError().add(requestErrors.getError());
    }

    // Get Record Request is done HERE!
    else {

      // Default is DC.
      RecordDelivererInterface recDeliv = this.recDelivDC;

      // Take IDIOM if IDIOM prefix.
      if (request.getMetadataPrefix().equals(TGConstants.METADATA_IDIOM_PREFIX)) {
        recDeliv = this.recDelivIDIOM;
      }

      // Finally start the QUERY!
      GetRecordType getRecord = new GetRecordType();

      String id = request.getIdentifier();
      getRecord = recDeliv.getRecordById(id);

      if (getRecord != null) {
        oai.setGetRecord(getRecord);
      } else {
        ErrorHandler idError = new ErrorHandler();
        idError.setError(TGConstants.OAI_NO_RECORD_MATCH, "The value of the identifier: "
            + request.getIdentifier() + " is unknown or illegal in this repository");
        oai.getError().add(idError.getError());
      }
    }

    return oaipmhRoot;
  }

  /**
   * checks the ListIdentifiers request of correctness and response including errors in case of an
   * incorrect request
   * 
   * @param obf
   * @param oai
   * @param request
   * @return
   * @throws ParseException
   */
  public JAXBElement<OAIPMHType> listIdentifiersRequest(ObjectFactory obf, OAIPMHType oai,
      RequestType request) throws ParseException {

    JAXBElement<OAIPMHType> oaipmhRoot = obf.createOAIPMH(oai);
    ErrorHandler requestErrors = IdentifierListDelivererAbstract.requestChecker(request);

    if (requestErrors.getError().getValue() != null) {
      oai.getError().add(requestErrors.getError());
    } else {

      // If metadataFormat IS SET, set recordListDeliverer accordingly.
      IdentifierListDelivererInterface idListDeliv = null;
      if (request.getMetadataPrefix() != null) {
        if (request.getMetadataPrefix().equals(TGConstants.METADATA_DC_PREFIX)) {
          idListDeliv = this.identifierListDC;
        } else {
          idListDeliv = this.identifierListIDIOM;
        }
      }

      // If metadata format IS NOT SET, check resumption token hash maps to decide if we have to
      // serve oai_dc or idiom_mets metadata format.
      else {
        boolean restokDCExisting = IdentifierListDelivererDC.cursorCollector != null
            && IdentifierListDelivererDC.cursorCollector.containsKey(request.getResumptionToken());
        boolean restokIDIOMExisting = IdentifierListDelivererIDIOM.cursorCollector != null
            && IdentifierListDelivererIDIOM.cursorCollector
                .containsKey(request.getResumptionToken());
        if (restokDCExisting) {
          idListDeliv = this.identifierListDC;
        } else if (restokIDIOMExisting) {
          idListDeliv = this.identifierListIDIOM;
        }
      }

      ListIdentifiersType listIdentifiers=null;
	try {
		listIdentifiers = idListDeliv.processIdentifierList(request.getFrom(),
		      request.getUntil(), request.getSet(), request.getResumptionToken());
	} catch (IOException e) {
		// TODO Auto-generated catch block
		e.printStackTrace();
	}

      if (listIdentifiers != null) {
          if(identifierListDC.getResultSize()==0) {
        	  requestErrors.setError("RecordMatchError", "The combination of the values of the from, until, set and metadataPrefix arguments results in an empty list.");
        	  oai.getError().add(requestErrors.getError());
          }else {
        oai.setListIdentifiers(listIdentifiers);
          }
      }
      //} else {
        /*ErrorHandler idError = new ErrorHandler();
        idError.setError(TGConstants.OAI_NO_RECORD_MATCH, "The value of the identifier: "
            + request.getIdentifier() + " is unknown or illegal in this repository");
        oai.getError().add(idError.getError());*/
      //}
    }

    return oaipmhRoot;
  }

  /**
   * checks the ListSets request of correctness and response including errors in case of an
   * incorrect request
   * 
   * @param obf
   * @param oai
   * @param request
   * @return
   */
  public JAXBElement<OAIPMHType> listSetsRequest(ObjectFactory obf, OAIPMHType oai,
      RequestType request) {

    ListSetsType sl = new ListSetsType();
    JAXBElement<OAIPMHType> oaipmhRoot = obf.createOAIPMH(oai);
    sl = this.setList.setListBuilder();
    oai.setListSets(sl);

    return oaipmhRoot;
  }

  /**
   * @param id
   * @param obf
   * @param oai
   * @param request
   * @return
   */
  public JAXBElement<OAIPMHType> listMetadataFormatsRequest(String id, ObjectFactory obf,
      OAIPMHType oai, RequestType request) {

    List<String> errorValues = new ArrayList<String>();
    ListMetadataFormatsType listMF = new ListMetadataFormatsType();
    JAXBElement<OAIPMHType> oaipmhRoot = obf.createOAIPMH(oai);

    if (this.metadataFormatList.requestChecker(request)) {

      if (id.isEmpty()) {
        listMF = this.metadataFormatList.setMetadataFormatListWithoutId();
      } else {
        listMF = this.metadataFormatList.setMetadataFormatList(request.getIdentifier());
      }

      if (listMF != null) {

        oai.setListMetadataFormats(listMF);
      } else {
        this.error.setError(TGConstants.OAI_NO_RECORD_MATCH, "The value of the identifier: "
            + request.getIdentifier() + " is unknown or illegal in this repository");
        oai.getError().add(this.error.getError());
      }
    } else {
      if (request.getFrom() != null) {
        errorValues.add("from");
      }
      if (request.getMetadataPrefix() != null) {
        errorValues.add("metadataPrefix");
      }
      if (request.getResumptionToken() != null) {
        errorValues.add("resumptionToken");
      }
      if (request.getSet() != null) {
        errorValues.add("set");
      }
      if (request.getUntil() != null) {
        errorValues.add("until");
      }

      this.error.setError(TGConstants.OAI_BAD_ARGUMENT, "The request includes illegal "
          + "arguments or is missing required arguments: " + errorValues);
      oai.getError().add(this.error.getError());
    }

    return oaipmhRoot;
  }

  /**
   * @param obf
   * @param oai
   * @param request
   * @return
   */
  public JAXBElement<OAIPMHType> listRecordsRequest(ObjectFactory obf, OAIPMHType oai,
      RequestType request) {

    // Create OAIPMH root object.
    JAXBElement<OAIPMHType> oaipmhRoot = obf.createOAIPMH(oai);

    // Check request's syntax.
    ErrorHandler requestErrors = RecordListDelivererAbstract.requestChecker(request);

    if (requestErrors.getError().getCode() != null) {
      oai.getError().add(requestErrors.getError());
    } else {

      // If metadataFormat IS SET, set recordListDeliverer accordingly.
      RecordListDelivererInterface recListDeliv = null;
      if (request.getMetadataPrefix() != null) {
        if (request.getMetadataPrefix().equals(TGConstants.METADATA_DC_PREFIX)) {
          recListDeliv = this.recordListDC;

        } else {
          recListDeliv = this.recordListIDIOM;
        }
      }

      // If metadata format IS NOT SET, check resumption token hash maps to decide if we have to
      // serve oai_dc or idiom_mets metadata format.
      else {
        boolean restokDCExisting = RecordListDelivererDC.cursorCollector != null
            && RecordListDelivererDC.cursorCollector.containsKey(request.getResumptionToken());
        boolean restokIDIOMExisting = RecordListDelivererIDIOM.cursorCollector != null
            && RecordListDelivererIDIOM.cursorCollector.containsKey(request.getResumptionToken());
        if (restokDCExisting) {
          recListDeliv = this.recordListDC;
        } else if (restokIDIOMExisting) {
          recListDeliv = this.recordListIDIOM;
        }
      }

      ListRecordsType listRecords = recListDeliv.getRecords(request.getFrom(), request.getUntil(),
          request.getSet(), request.getResumptionToken());

      if (listRecords != null ) {
          if(recordListDC.getResultSize()==0) {
        	  requestErrors.setError("RecordMatchError", "The combination of the values of the from, until, set and metadataPrefix arguments results in an empty list.");
        	  oai.getError().add(requestErrors.getError());
          }else {
        	  oai.setListRecords(listRecords);	  
          }
        
      }
    }

    return oaipmhRoot;
  }

  /**
   * Taking the sting from the REST request and converting into an OAIPMH verb type
   * 
   * @param verb : what is to within the repository
   * @return verbParam: the verbParam as an OAIPMH verb type
   */
  public VerbType setVerb(String verb) {

    VerbType verbParam = null;

    if (verb == null) {
      verb = "Identify";
    }

    if (verb != null && verb.equals("Identify")) {
      verbParam = VerbType.IDENTIFY;
    } else if (verb != null && verb.equals("ListMetadataFormats")) {
      verbParam = VerbType.LIST_METADATA_FORMATS;
    } else if (verb != null && verb.equals("ListSets")) {
      verbParam = VerbType.LIST_SETS;
    } else if (verb != null && verb.equals("ListIdentifiers")) {
      verbParam = VerbType.LIST_IDENTIFIERS;
    } else if (verb != null && verb.equals("ListRecords")) {
      verbParam = VerbType.LIST_RECORDS;
    } else if (verb != null && verb.equals("GetRecord")) {
      verbParam = VerbType.GET_RECORD;
    } else {
      // FIXME
      this.error.setError("VerbError", "Illegal OAI verb");
    }

    return verbParam;
  }

  /**
   * Setting the from value for the request
   * 
   * @param from
   */

  public void setFromRequestValue(String from, RequestType request) {
    if (!from.isEmpty()) {
      request.setFrom(from);
    }
  }

  /**
   * Setting the until value for the request
   * 
   * @param until
   */
  public void setUntilRequestValue(String until, RequestType request) {
    if (!until.isEmpty()) {
      request.setUntil(until);
    }
  }

  /**
   * Setting the identifier value for the request
   * 
   * @param identifier
   */
  public void setIdentifierRequestValue(String identifier, RequestType request) {
    if (!identifier.isEmpty()) {
      request.setIdentifier(identifier);
    }
  }

  /**
   * Setting the metadataPrefix value for the request
   * 
   * @param metadataPrefix
   */
  public void setMetadataPrefixRequestValue(String metadataPrefix, RequestType request) {
    if (!metadataPrefix.isEmpty()) {
      request.setMetadataPrefix(metadataPrefix);
    }
  }

  /**
   * Setting the resumptionToken value for the request
   * 
   * @param resumptionToken
   */
  public void setResumptionTokenRequestValue(String resumptionToken, RequestType request) {
    if (!resumptionToken.isEmpty()) {
      request.setResumptionToken(resumptionToken);
    }
  }

  /**
   * <p>
   * Setting the set value for the request
   * </p>
   * 
   * @param set
   * @param request
   */
  public void setSetRequestValue(String set, RequestType request) {
    if (!set.isEmpty()) {
      request.setSet(set);
    }
  }

  /**
   * @param textgrid
   */
  public void setTextGrid(boolean textgrid) {
    this.textgrid = textgrid;
  }

  /**
   * @param dariah
   */
  public void setDariah(boolean dariah) {
    this.dariah = dariah;
  }

  // **
  // PRIVATE METHODS
  // **

  /**
   * @param verb
   * @param identifier
   * @param metadataPrefix
   * @param set
   * @param from
   * @param until
   * @param resumptionToken
   * @return
   */
  private JAXBElement<OAIPMHType> handleRequest(String verb, String identifier,
      String metadataPrefix, String set, String from, String until, String resumptionToken) {

    ObjectFactory obf = new ObjectFactory();
    OAIPMHType response = new OAIPMHType();
    JAXBElement<OAIPMHType> oaipmhRoot = obf.createOAIPMH(response);
    VerbType verbParam = setVerb(verb);
    RequestType request = new RequestType();

    /*----Setting the responseDate of today-----*/
    try {
      XMLGregorianCalendar convertedDateFormat;
      convertedDateFormat = OAIPMHUtillities
          .convertDateFormat(OAIPMHUtillities.getXMLGregorianCalendarNow().toString());
      response.setResponseDate(convertedDateFormat);
    } catch (ParseException e) {
      this.log.debug("can not parse date format");
    } catch (DatatypeConfigurationException e) {
      this.log.debug("Datatype configuration failed");
    }

    /* converting the request values */

    setFromRequestValue(from, request);
    setUntilRequestValue(until, request);
    setIdentifierRequestValue(identifier, request);
    setMetadataPrefixRequestValue(metadataPrefix, request);
    setResumptionTokenRequestValue(resumptionToken, request);
    setSetRequestValue(set, request);
    request.setVerb(verbParam);

    if (this.textgrid == true) {
      request.setValue(TGConstants.TG_REP_BASEURL);
    }
    if (this.dariah == true) {
      request.setValue(DARIAHConstants.DARIAH_REP_BASEURL);
    }

    if (verbParam != null) {
      if (verbParam.value().equals("Identify")) {
        oaipmhRoot = identifyRequest(obf, response, request);
      } else if (verbParam.value().equals("GetRecord")) {
        oaipmhRoot = getRecordRequest(obf, response, request);
      } else if (verbParam.value().equals("ListIdentifiers")) {
        try {
          oaipmhRoot = listIdentifiersRequest(obf, response, request);
        } catch (ParseException e) {
          // TODO Auto-generated catch block
          e.printStackTrace();
        }
      } else if (verbParam.value().equals("ListSets")) {
        oaipmhRoot = listSetsRequest(obf, response, request);
      } else if (verbParam.value().equals("ListMetadataFormats")) {
        oaipmhRoot = listMetadataFormatsRequest(identifier, obf, response, request);
      } else if (verbParam.value().equals("ListRecords")) {
        oaipmhRoot = listRecordsRequest(obf, response, request);
      }
    } else {
      this.error.setError("VerbError", "Illegal OAI verb");
      response.getError().add(this.error.getError());
    }

    response.setRequest(request);

    oaipmhRoot.setValue(response);

    return oaipmhRoot;
  }

}
